/*
 * This file is part of the flashrom project.
 *
 * Copyright (C) 2023 Celliwig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/*
 * Routines to access (emulated) Intel 8042 keyboard controller
 * Needed because SuperIO's KBC can be become unstable when they are combined
 * with flash write functions.
 */

#include <strings.h>
#include <string.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#if defined(__i386__) || defined(__x86_64__)

#include "hwaccess_x86_io.h"

#include "i8042.h"

// Unbind module from hardware
uint8_t i8042_kbc_driver_unbind()
{
	FILE *sysfs;

	sysfs = fopen(I8042_SYSFS_PATH "unbind", "w");
	if (sysfs == NULL) return -1;
	if (fprintf(sysfs, "i8042") < 0) return -2;
	fclose(sysfs);

	return 0;
}

// Rebind module to hardware
uint8_t i8042_kbc_driver_rebind()
{
	FILE *sysfs;

	sysfs = fopen(I8042_SYSFS_PATH "bind", "w");
	if (sysfs == NULL) return -1;
	if (fprintf(sysfs, "i8042") < 0) return -2;
	fclose(sysfs);

	return 0;
}

// Send keyboard controller a command (timeout=0 for no timeout)
uint8_t i8042_kbc_cmd_timeout(uint8_t cmd, uint32_t timeout)
{
	// Read KBC input buffer status (wait for empty)
	while (0x2 & INB(I8042_KBC_STATUS_PORT)) {
		OUTB(0, I8042_UNDEFINED_PORT);
		if (timeout > 0) {
			timeout--;
			if (timeout == 0) return -1;
		}
	}
	OUTB(cmd, I8042_KBC_COMMAND_PORT);
	return 0;
}

// Send keyboard controller a command (no timeout)
void i8042_kbc_cmd(uint8_t cmd)
{
	i8042_kbc_cmd_timeout(cmd, 0);
}

// Get keyboard controller data (timeout=0 for no timeout)
uint8_t i8042_kbc_data_read_timeout(uint8_t *buf, uint32_t timeout)
{
	// Read KBC output buffer status (wait for full)
	while (!(0x1 & INB(I8042_KBC_STATUS_PORT))) {
		OUTB(0, I8042_UNDEFINED_PORT);
		if (timeout > 0) {
			timeout--;
			if (timeout == 0) return -1;
		}
	}
	*buf = INB(I8042_KBC_DATA_PORT);
	return 0;
}

// Get keyboard controller data (no timeout)
uint8_t i8042_kbc_data_read()
{
	uint8_t val;
	i8042_kbc_data_read_timeout(&val, 0);
	return val;
}

// Set keyboard controller data (timeout=0 for no timeout)
uint8_t i8042_kbc_data_write_timeout(uint8_t val, uint32_t timeout)
{
	// Read KBC input buffer status (wait for empty)
	while (0x2 & INB(I8042_KBC_STATUS_PORT)) {
		OUTB(0, I8042_UNDEFINED_PORT);
		if (timeout > 0) {
			timeout--;
			if (timeout == 0) return -1;
		}
	}
	OUTB(val, I8042_KBC_DATA_PORT);
	return 0;
}

// Set keyboard controller data (no timeout)
void i8042_kbc_data_write(uint8_t val)
{
	i8042_kbc_data_write_timeout(val, 0);
}

// Get PS/2 controller configuration
uint8_t i8042_kbc_cfg_get()
{
	// Get controller configuration byte
	i8042_kbc_cmd(0x20);
	return i8042_kbc_data_read();
}

// Set PS/2 controller configuration
void i8042_kbc_cfg_set(uint8_t cfg)
{
	// Set controller configuration byte
	i8042_kbc_cmd(0x60);
	i8042_kbc_data_write(cfg);
}

// Send reset command to specified PS/2 device
uint8_t i8042_kbc_device_reset(uint8_t device)
{
	uint8_t val;
	bool read_extra = false;

	if (device == 2) {
		// Write following device command to port 2
		i8042_kbc_cmd(0xd4);
		// It's a mouse, so get device ID
		read_extra = true;
	}
	// Reset selected device
	i8042_kbc_data_write(0xff);

	if (i8042_kbc_data_read_timeout(&val, 0xffffff)) return -1;
	if (val != 0xfa) return -2;
	if (i8042_kbc_data_read_timeout(&val, 0xffffff)) return -3;
	if (val != 0xaa) return -4;
	if (read_extra) {
		if (i8042_kbc_data_read_timeout(&val, 0xffffff)) return -5;
	}

	return 0;
}

// Send enable/disable command to specified PS/2 device
uint8_t i8042_kbc_device_enable(uint8_t device, bool enable)
{
	uint8_t val;

	if (device == 2) {
		// Write following device command to port 2
		i8042_kbc_cmd(0xd4);
	}
	if (enable) {
		// Enable selected device
		i8042_kbc_data_write(0xf4);
	} else {
		// Disable selected device
		i8042_kbc_data_write(0xf5);
	}
	if (i8042_kbc_data_read_timeout(&val, 0xffffff)) return -1;
	if (val != 0xfa) return -2;

	return 0;
}

#endif
